(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

Quiet[
    BeginPackage["AtomicDensityMatrix`WignerEckart`", {
        "ProcessOptions`",
        "Units`",
        "PhysicalConstants`",
        "AtomicDensityMatrix`"
    }],
    General::obspkg
];

MatrixElementMask::usage = "";

Begin["`Private`"];

MatrixElementMask[sys1_?AtomicSystemQ, sys2_?AtomicSystemQ, allowedCouplings_, neglectedCouplings_] :=
    Boole@TransitionQ[sys1, sys2, allowedCouplings] * (1 - Boole@TransitionQ[sys1, sys2, neglectedCouplings]);

Options@WignerEckart = {Representation -> Automatic, AllowedCouplings -> All, NeglectedCouplings -> None};

OptionProcessor[AllowedCouplings] = (# /. Automatic :> OptionValue[WignerEckart, AllowedCouplings] /. {All -> True, None -> False}&);
OptionProcessor[NeglectedCouplings] = (# /. Automatic :> OptionValue[WignerEckart, NeglectedCouplings] /. {All -> True, None -> False}&);

WignerEckart[sys_?AtomicSystemQ, op_, opts : OptionsPattern[]] := WignerEckart[sys, op, sys, opts];

WignerEckart[s1_AtomicState, {a_, k_Integer}, s2_AtomicState, opts : OptionsPattern[]] :=
    WignerEckart[{s1}, {a, k}, {s2}, opts][[All, All, 1, 1]];

WignerEckart[s1_AtomicState, {a_, k_Integer, q : _Integer | {__Integer}}, s2_AtomicState, opts : OptionsPattern[]] :=
    WignerEckart[{s1}, {a, k, q}, {s2}, opts][[1, 1]];

WignerEckart[sys1_?AtomicSystemQ, {a_, k_Integer}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] :=
    WignerEckart[sys1, {a, k, Range[k, -k, -1]}, sys2, opts];

WignerEckart[sys1_?AtomicSystemQ, {a_, k_Integer, q_Integer}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] :=
    WignerEckart[sys1, {a, k, {q}}, sys2, opts][[1, 1]];

WignerEckart[label_, j_?HalfIntegerQ, a_, opts : OptionsPattern[]] := WignerEckart[ZeemanSystem[label, j], a, opts];

WignerEckart[j_?HalfIntegerQ, a_, opts : OptionsPattern[]] := WignerEckart[ZeemanSystem[j], a, opts];

(WignerEckart[sys1_?AtomicSystemQ, {a_, k_Integer, {__Integer}}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] /;
    MatchQ[AtomicSystemType@Join[sys1, sys2], "Toy"] :=
        Module[{allowedCouplings, neglectedCouplings},
            Normal@{{MatrixElementMask[sys1, sys2, allowedCouplings, neglectedCouplings] ReducedME[sys1, {a, k}, sys2]}}
        ]
) // ProcessOptions;

(WignerEckart[sys1_?AtomicSystemQ, {a_, k_Integer, {__Integer}}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] /;
    MatchQ[AtomicSystemType@Join[sys1, sys2], "Hyperfine"] :=
    Module[{allowedCouplings, neglectedCouplings},
        Normal@{{MatrixElementMask[sys1, sys2, allowedCouplings, neglectedCouplings] ReducedME[sys1, {a, k}, sys2]/If[k == 0, Sqrt[2F[sys1] + 1], 1]}}
    ]
) // ProcessOptions;

(WignerEckart[sys1_?AtomicSystemQ, {a_, k_Integer, qs : {__Integer}}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {representation},
    Switch[representation,
        {"Zeeman"},
        WignerEckartZ[sys1, {a, k, qs}, sys2, FilterRules@opts],
        {"PolarizationMoments", ___},
        WignerEckartPM[sys1, {a, k, qs}, sys2, Rest@representation, FilterRules@opts]
    ]
]) // ProcessOptions;

Options[WignerEckartZ] = FilterRules[Options[WignerEckart], Except@Representation];

(WignerEckartZ[sys1_?AtomicSystemQ, {a_, k_Integer, qs : {__Integer}}, sys2_?AtomicSystemQ, OptionsPattern[]] := Module[
    {allowedCouplings, neglectedCouplings, fm1, fm2, rme, mask},
    {fm1, fm2} = Transpose /@ Through /@ {F, M} /@ {sys1, sys2};
    rme = ReducedME[sys1, {a, k}, sys2];
    mask = MatrixElementMask[sys1, sys2, allowedCouplings, neglectedCouplings];
    Expand[{rme wignerEckartFactor[fm1, {k, #}, fm2, mask] & /@ qs}]
]) // ProcessOptions;

wignerEckartFactor[fm1 : {{_, _}..}, {k_, q_}, fm2 : {{_, _}..}, mask_] := wignerEckartFactor[fm1, {k, q}, fm2, mask] =
    MapThread[wignerEckartFactor[#1[[1]], {k, q}, #1[[2]], #2]&, {DirectProduct[List, fm1, fm2], mask}, 2];

wignerEckartFactor[{_, _}, {_, _}, {_, _}, 0] = 0;

wignerEckartFactor[{f1_, m1_}, {k_, q_}, {f2_, m2_}, 1] /;
    m2 + q != m1 || Abs[f2 - k] > f1 || f2 + k < f1 || Abs[m2] > f2 || Abs[q] > k || Abs[m1] > f1 := 0;

wignerEckartFactor[{f1_, m1_}, {k_, q_}, {f2_, m2_}, 1] := wignerEckartFactor[{f1, m1}, {k, q}, {f2, m2}] =
    (2 f1 + 1)^(-1 / 2) CG[{f2, m2}, {k, q}, {f1, m1}];

Options[WignerEckartPM] = Join[{MaxRanks -> Automatic}, FilterRules[Options[WignerEckart], Except@Representation]];

(WignerEckartPM[sys1_?AtomicSystemQ, {a_, k_Integer, qs : {__Integer}}, sys2_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {allowedCouplings, neglectedCouplings, reducedSys1, reducedSys2, rme},
    {reducedSys1, reducedSys2} = RemoveZeemanSublevels /@ {sys1, sys2};
    rme = MatrixElementMask[reducedSys1, reducedSys2, allowedCouplings, neglectedCouplings] ReducedME[reducedSys1, {a, k}, reducedSys2];
    Normal@{rme / Sqrt[2k + 1] #& /@ PolarizationMoment[reducedSys1, reducedSys2, All, {k, qs}, FilterRules@opts]}
]) // ProcessOptions;

(*Reduced matrix elements*)

ReducedME[sys_?AtomicSystemQ, {op_, k_}] := ReducedME[sys, {op, k}, sys];

ReducedME[sys1_?AtomicSystemQ, {op_, k_}, sys2_?AtomicSystemQ] := ArrayFlatten[
    DirectProduct[ConstantArray[1, {##}]&, CountZeemanSublevels@sys1, CountZeemanSublevels@sys2] *
        DirectProduct[ReducedME[#1, {op, k}, #2] &, RemoveZeemanSublevels@sys1, RemoveZeemanSublevels@sys2]
];

(*Polarization operator*)

ReducedME[s1_AtomicState, {Polarization, k_Integer}, s2_AtomicState] := Boole[StateLabel@s1 === StateLabel@s2] Sqrt[2k + 1];

ReducedME[s1_AtomicState, {Polarization[{s1p_, f1_?NumericQ}, {s2p_, f2_?NumericQ}], k_Integer}, s2_AtomicState] :=
    Boole[{StateLabel@s1, StateLabel@s2, F@s1, F@s2} === {s1p, s2p, f1, f2}] Sqrt[2k + 1];

ReducedME[s1_AtomicState, {Polarization[s1p_, s2p_], k_Integer}, s2_AtomicState] :=
    Boole[{StateLabel@s1, StateLabel@s2} === {s1p, s2p}] Sqrt[2k + 1];

(*Hyperfine*)

(*Special case, statistical weight operator*)

ReducedME[s1_AtomicState, {StatisticalWeight, 0}, s2_AtomicState] /; HyperfineTypeSystem[s1, s2] :=
    Module[{j1, i1, f1, j2, i2, f2, sJ1, sJ2},
        {{i1, j1, f1}, {i2, j2, f2}} = Through /@ {NuclearSpin, J, F} /@ {s1, s2};
        {sJ1, sJ2} = DeleteParameter[{s1, s2}, F, NuclearSpin];
        (-1)^(j1 + i1 + f2) Sqrt[(2 f1 + 1)(2 f2 + 1)] / (2 i1 + 1) SJS[{j1, f1, i1}, {f2, j2, 0}] ReducedME[sJ1, {StatisticalWeight, 0}, sJ2]
    ];

(*The energy of a given hyperfine level*)

hyperfineK[j_, i_, f_] = f(f + 1) - i(i + 1) - j(j + 1);

EvaluateHyperfineShift[expr_] := expr /. HyperfineShift[j_, i_, f_, a_, b_] :> With[
    {k = hyperfineK[j, i, f]},
    1/2 a k + Piecewise[{{
        b (3/2 k (k + 1) - 2 i (i + 1) j (j + 1)) / (2 i (2 i - 1) 2 j (2 j - 1)),
        j > 1/2 && i > 1/2
    }}]
];

HyperfineShift[s_AtomicState] /; HyperfineTypeSystem@s :=
    EvaluateHyperfineShift@HyperfineShift[J@s, NuclearSpin@s, F@s, HyperfineA@s, HyperfineB@s];

HyperfineShift[s_AtomicState] := 0;

HyperfineShift[s_?AtomicSystemQ] := HyperfineShift /@ s;

hyperfineRMEFactor[s1_, k_Integer, s2_] := hyperfineRMEFactor[s1, k, s2] =
    (-1)^(J@s1 + NuclearSpin@s1 + F@s2 + k) Sqrt[(2F@s1 + 1)(2F@s2 + 1)] *
        SJS[{J@s1, F@s1, NuclearSpin@s1}, {F@s2, J@s2, k}];

ReducedME[s1_AtomicState, {Energy, 0}, s2_AtomicState] /; HyperfineTypeSystem[s1, s2] :=
    Module[{sJ1, sJ2},
        {sJ1, sJ2} = DeleteParameter[{s1, s2}, F, NuclearSpin];
        hyperfineRMEFactor[s1, 0, s2] *
            (ReducedME[sJ1, {Energy, 0}, sJ2] + HyperfineShift@s1 ReducedME[sJ1, {Identity, 0}, sJ2])
    ];

(* ReducedME for hyperfine structure, assuming the operator commutes with nuclear spin *)
(* (I J_1 F_1||a^k||I J_2 F_2) = (-1)^(J_1+I+F_2+k) Sqrt[(2 F_1+1)(2 F_2+1)] SJS[{J_1,F_1,I},{F_2,J_2,k}] (J_1||a^k||J_2) *)
(* Sobelman Eq. (4.175), p. 84 *)

ReducedME[s1_AtomicState, {a_, k_Integer}, s2_AtomicState] /; HyperfineTypeSystem[s1, s2] :=
    ReducedME[s1, {a, k}, s2] =
        hyperfineRMEFactor[s1, k, s2] *
            ReducedME[DeleteParameter[s1, F, NuclearSpin], {a, k}, DeleteParameter[s2, F, NuclearSpin]];

(*Zero*)

ReducedME[s1_AtomicState, {Zero, 0}, s2_AtomicState] = 0;

(*Identity operator*)

(*[Var88, Sec. 13.1.1, Eq. (3)]*)

ReducedME[s1_AtomicState, {Identity, 0}, s2_AtomicState] :=
    Boole[StateLabel@s1 === StateLabel@s2] Sqrt[2J@s1 + 1];

(*Ground state operator*)

ReducedME[s1_AtomicState, {GroundState, 0}, s2_AtomicState] :=
    Boole@GroundStateQ@s1 ReducedME[s1, {Identity, 0}, s2];

(*Excited state operator*)

ReducedME[s1_AtomicState, {ExcitedState, 0}, s2_AtomicState] :=
    Boole@Not@GroundStateQ@s1 ReducedME[s1, {Identity, 0}, s2];

(*Natural width operator*)

ReducedME[s1_AtomicState, {NaturalWidth, 0}, s2_AtomicState] :=
    NaturalWidth@s1 ReducedME[s1, {Identity, 0}, s2];

(*Statistical weight operator*)

ReducedME[s1_AtomicState, {StatisticalWeight, 0}, s2_AtomicState] :=
    1 / (2J@s1 + 1) ReducedME[s1, {Identity, 0}, s2];

(*Unperturbed energy operator*)

ReducedME[s1_AtomicState, {Energy, 0}, s2_AtomicState] :=
    Energy@s1 ReducedME[s1, {Identity, 0}, s2];

(*E1 dipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

ReducedME[s1_AtomicState, {Dipole, 1}, s2_AtomicState] /;
    Parity@s1 === Parity@s2 || StateLabel@s1 === StateLabel@s2 :=
    ReducedME[s1, {Dipole, 1}, s2] = 0;

(*Reorder states, using (\[Gamma] J||d||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||d||\[Gamma] J) *)
(*If states are in the right order, replace expression with a placeholder whose value can be set later*)

ReducedME[s1_AtomicState, {Dipole, 1}, s2_AtomicState] :=
    ReducedME[s1, {Dipole, 1}, s2] =
        With[{lab1 = StateLabel@s1, lab2 = StateLabel@s2},
            If[OrderedQ@{lab1, lab2},
                ReducedME[lab1, {Dipole, 1}, lab2],
                If[MatchQ[AtomicSystemType@s1, "Fine" | "Zeeman"], (-1)^(J@s1 - J@s2), 1] ReducedME[lab2, {Dipole, 1}, lab1]
            ]
        ];

(*Dipole operator with RME set to unity*)

ReducedME[s1_AtomicState, {"NormalizedDipole", 1}, s2_AtomicState] /;
    Parity@s1 === Parity@s2 || StateLabel@s1 === StateLabel@s2 :=
    ReducedME[s1, {"NormalizedDipole", 1}, s2] = 0;

ReducedME[s1_AtomicState, {"NormalizedDipole", 1}, s2_AtomicState] :=
    ReducedME[s1, {"NormalizedDipole", 1}, s2] =
        With[{lab1 = StateLabel@s1, lab2 = StateLabel@s2},
            If[OrderedQ@{lab1, lab2},
                1,
                If[MatchQ[AtomicSystemType@s1, "Fine" | "Zeeman"], (-1)^(J@s1 - J@s2), 1]
            ]
        ];

DipoleExpansion[state1_AtomicState, state2_AtomicState] := Module[
    {l1, j1, l2, j2, lab1, lab2, w1, w2, gam1, gam2, br12, br21},
    {{l1, j1, w1, lab1, gam1}, {l2, j2, w2, lab2, gam2}} = Through /@ {L, J, Energy, StateLabel, NaturalWidth} /@ {state1, state2};
    {br12, br21} = {BranchingRatio[state1][state2], BranchingRatio[state2][state1]};
    ReducedME[lab1, {Dipole, 1}, lab2] ->
        Which[
            br21 === 0 && br12 =!= 0,
            Sqrt[3]/2 Sqrt[(2 j2 + 1) gam2 br12 / (w2 - w1)^3],
            br12 === 0 && br21 =!= 0,
            Sqrt[3]/2 Sqrt[(2 j1 + 1) gam1 br21 / (w1 - w2)^3],
            True,
            ReducedME[lab1, {Dipole, 1}, lab2]
        ]
];

ExpandDipoleRME[sys_, expr_] :=
    expr /. Flatten@DirectSelfProduct[DipoleExpansion, Union[sys, SameTest -> (StateLabel@#1 === StateLabel@#2&)]];

ReducedRabiFrequency[sys_, stateLabel1_ -> stateLabel2_, intensity_, "SI"] := Module[{dADM, dSI, eSI},
    dADM = ExpandDipoleRME[sys, ReducedME[stateLabel1, {Dipole, 1}, stateLabel2]];
    dSI = dADM Sqrt[PlanckConstantReduced SpeedOfLight^3] Sqrt[4 Pi VacuumPermittivity];
    eSI = Sqrt[2 intensity / (SpeedOfLight VacuumPermittivity)];
    PowerExpand[dSI eSI/PlanckConstantReduced /. Joule -> Watt Second]
];

(*Spontaneous emission*)
(*
ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; StateLabel[s1] === StateLabel[s2] = 0;

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; Parity[s1] === Parity[s2] = 0;

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; !OrderedQ[{StateLabel[s1], StateLabel[s2]}] :=
    If[AtomicSystemType[s1] === "Toy", 1, (-1)^(J[s1] - J[s2])] ReducedME[s2, {SpontaneousEmission, 1}, s1];

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] :=
    If[LevelPosition[s1] === Lower,
        Sqrt[(2J[s2] + 1)NaturalWidth[s2]BranchingRatio[StateLabel[s1]][s2]],
        Sqrt[(2J[s1] + 1)NaturalWidth[s1]BranchingRatio[StateLabel[s2]][s1]]
    ];
*)

(*Angular momentum operator*)

ReducedME[s1_AtomicState, {J, 1}, s2_AtomicState] :=
    Sqrt[J@s1(J@s1 + 1)(2J@s1 + 1)] Boole[StateLabel@s1 === StateLabel@s2];

ReducedME[state1_AtomicState, {S, 1}, state2_AtomicState] := Module[
    {s1, l1, j1, s2, l2, j2, lab1, lab2},
    {{s1, l1, j1, lab1}, {s2, l2, j2, lab2}} = Through /@ {S, L, J, StateLabel} /@ {state1, state2};
    Which[
        (*Coupling within the same J-state, evaluate explicitly*)
        lab1 === lab2,
        If[j1 < 1 / 2,
            0,
            Sqrt[j1 (j1 + 1) (2 j1 + 1)] (j1 (j1 + 1) + s1 (s1 + 1) - l1 (l1 + 1)) / (2 j1 (j1 + 1))
        ],
        (*Coupling between fine-structure states*)
        FineStructureStatesQ[state1, state2],
        If[DefinedQ[{state1, state2}, {S, L, J}],
            (*If all quantum numbers are defined, evaluate explicitly*)
            If[s1 =!= s2 || l1 =!= l2,
                0,
                (-1)^(s1 + l1 + j2 + 1) Sqrt[s1 (s1 + 1) (2 s1 + 1) (2 j1 + 1) (2 j2 + 1)] SJS[{s1, j1, l1}, {j2, s1, 1}]
            ],
            (*Otherwise return symbolic placeholder*)
            ReducedME[lab1, {S, 1}, lab2]
        ],
        (*If coupling is not between fine-structure states, return zero*)
        True,
        0
    ]
];

ReducedME[state1_AtomicState, {L, 1}, state2_AtomicState] := Module[
    {s1, l1, j1, s2, l2, j2, lab1, lab2},
    {{s1, l1, j1, lab1}, {s2, l2, j2, lab2}} = Through /@ {S, L, J, StateLabel} /@ {state1, state2};
    Which[
        (* Coupling within the same J-state, evaluate explicitly. *)
        lab1 === lab2,
        If[j1 < 1 / 2,
            0,
            Sqrt[j1 (j1 + 1) (2 j1 + 1)] (j1 (j1 + 1) + l1 (l1 + 1) - s1 (s1 + 1)) / (2 j1 (j1 + 1))
        ],
        (* Coupling between fine-structure states. *)
        FineStructureStatesQ[state1, state2],
        If[DefinedQ[{state1, state2}, {S, L, J}],
            (* If all quantum numbers are defined, evaluate explicitly. *)
            If[s1 =!= s2 || l1 =!= l2,
                0,
                (-1)^(s1 + l1 + j2 + 1) Sqrt[l1 (l1 + 1) (2 l1 + 1) (2 j1 + 1) (2 j2 + 1)] SJS[{l1, j1, s1}, {j2, l1, 1}]
            ],
            (* Otherwise, return symbolic placeholder. *)
            ReducedME[lab1, {L, 1}, lab2]
        ],
        (* If coupling is not between fine-structure states, return zero. *)
        True,
        0
    ]
];

(*Magnetic moment operator*)

(*g factors*)

LandeGJ[0, l_, s_] = 0;
LandeGJ[j_, l_, s_] = 1 + (j (j + 1) + s (s + 1) - l (l + 1)) / (2 j (j + 1));
LandeGJ[s_AtomicState] := LandeGJ[J@s, L@s, S@s];
SetAttributes[LandeGJ, Listable];

LandeGF[0, j_, l_, s_, i_] = 0;
LandeGF[f_, j_, l_, s_, i_] = LandeGJ[j, l, s] (f (f + 1) + j (j + 1) - i (i + 1)) / (2 f (f + 1));
LandeGF[s_AtomicState] := LandeGF[F@s, J@s, L@s, S@s, NuclearSpin@s];
SetAttributes[LandeGF, Listable];

ReducedME[s1_AtomicState, {MagneticMoment, 1}, s2_AtomicState] :=
    -BohrMagneton Which[
        (* For coupling within the same state, calculate using g-factor for efficiency. *)
        StateLabel@s1 === StateLabel@s2,
        Which[
            DefinedQ[s1, GFactor],
            GFactor@s1,
            DefinedQ[s1, {J, L, S}],
            LandeGJ@s1,
            True,
            GFactor@s1
        ] ReducedME[s1, {J, 1}, s2],
        (* Between two fine-structure states, calculate using L and S. *)
        FineStructureStatesQ[s1, s2],
        ReducedME[s1, {L, 1}, s2] + 2 ReducedME[s1, {S, 1}, s2],
        (* Otherwise, return zero. *)
        True,
        0
    ];

(*MagneticDipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /;
	Parity[s1]!=Parity[s2]||StateLabel[s1]==StateLabel[s2] := 0*)

(*Reorder states, using (s1 J||MagneticDipole||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||MagneticDipole||\[Gamma] J) *)
(* NEED TO CHECK IF THIS IS CORRECT!*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /; 
	!OrderedQ[{StateLabel[s1],StateLabel[s2]}] :=
		(-1)^(J[s1]-J[s2]) ReducedME[s2,{MagneticDipole,1},s1]*)

(*Polarizability operator*)

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[\!\(TraditionalForm\`*)
(*\*SubscriptBox[\(H\), \(Stark\)] = {\**)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"]\[CircleTimes]\**)
(*SuperscriptBox[*)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], "*"]}\), \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

alphaFactor[k_Integer, j_?HalfIntegerQ] /; k >= 2j + 1 := 0;

alphaFactor[k_Integer, j_?HalfIntegerQ] := alphaFactor[k, j] =
    Simplify[
        -(1/2)(Sqrt[2 j + 1]) / (CG[{1, Mod[k, 2]}, {1, -Mod[k, 2]}, {k, 0}] CG[{j, j}, {k, 0}, {j, j}]),
        Element[2 j, Integers]
    ];

ReducedME[s1_AtomicState, {Polarizability, k_Integer}, s2_AtomicState] :=
    alphaFactor[k, J@s1] Polarizability[k][s1] Boole[StateLabel@s1 === StateLabel@s2];

(*TransitionPolarizability operator*)

(*ReducedME[s1_AtomicState, {TransitionPolarizability, k_Integer}, s2_AtomicState] :=*)
(*Switch[k, 0, -Sqrt[3], 1, Sqrt[2], 2, 1]TransitionPolarizability[k][s1, s2] **)
(*(1 - Boole[StateLabel[s1] === StateLabel[s2]]);*)

End[];

EndPackage[];
